"""
The builder aims to generate crystal structure from the defined
building blocks (e.g., SiO2 with 4-coordined Si and 2-coordinated O)
1. Generate all possible wp combinations
2. For each wp combination,

    2.1 generate structure randomly
    2.2 optimize the geomtry
    2.3 parse the coordination
    2.4 save the qualified structure to ase.db

Todo:
1. Symmetry support for dSdX
2. add parallel for gulp optimiza
"""

# Standard Libraries
import os
import numpy as np
from scipy.optimize import minimize
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import Pool
from collections import deque

# Material science Libraries
from ase.db import connect
from pyxtal import pyxtal
from pyxtal.util import generate_wp_lib
from pyxtal.lattice import Lattice
from pyxtal.db import database_topology
from pyxtal.lego.util import get_input_from_letters, calculate_S, calculate_dSdx
from pyxtal.lego.basinhopping import basinhopping
from pyxtal.lego.SO3 import SO3

# Logging and debugging
import logging
from time import time
# np.set_printoptions(precision=3, suppress=True)

VECTORS = np.array([[x1, y1, z1] for x1 in range(-1, 2) for y1 in range(-1, 2) for z1 in range(-1, 2)])

def generate_wp_lib_par(spgs, composition, num_wp, num_fu, num_dof):
    """
    A wrapper to generate the wp list in parallel
    """
    my_spgs, wp_libs = [], []
    for spg in spgs:
        wp_lib = generate_wp_lib([spg], composition, num_wp, num_fu, num_dof),
        if len(wp_lib) > 0:
            wp_libs.append(wp_lib)
            my_spgs.append(spg)
    return (my_spgs, wp_libs)

def generate_xtal_par(wp_libs, niter, dim, elements, calculator, ref_environments,
                      criteria, T, N_max, early_quit):
    """
    A wrapper to call generate_xtal function in parallel
    """
    xtals, sims = [], []
    for wp_lib in wp_libs:
        (number, spg, wps, dof) = wp_lib
        xtal, sim = generate_xtal(dim, spg, wps, niter*dof, elements, calculator,
                                  ref_environments, criteria, T, N_max, early_quit)
        if xtal is not None:
            xtals.append(xtal)
            sims.append(sim)

    return (xtals, sims)

def minimize_from_x_par(*args):
    """
    A wrapper to call minimize_from_x function in parallel
    """
    dim, wp_libs, elements, calculator, ref_environments, opt_type, T, niter, early_quit, minimizers = args[0]
    xtals = []
    xs = []
    for wp_lib in wp_libs:
        (x, spg, wps) = wp_lib
        res = minimize_from_x(x, dim, spg, wps, elements, calculator,
                              ref_environments, T, niter,
                              early_quit, opt_type, minimizers,
                              filename=None)
        if res is not None:
            xtals.append(res[0])
            xs.append(res[1])
    return xtals, xs


def generate_xtal(dim, spg, wps, niter, elements, calculator,
                  ref_environments, criteria, T, N_max, early_quit,
                  dump=False, random_state=None, verbose=False):
    """
    Generate a crystal with the desired local environment

    Args:
        dim (int): 1, 2, 3
        spg: pyxtal.symmetry.Group object
        wps: list of wps for the disired crystal (e.g., [wp1, wp2])
        ref_env: reference enviroment
        f: callable function to compute env
        n_iter (int):
        T (float): for basinhopping

    Returns:
        xtal and its similarity
    """

    # Here we start to optimize the xtal based on similarity

    print("\n", dim, spg, wps, T, N_max, early_quit)
    count = 0
    while True:

        filename = 'opt.txt' if dump else None
        result = minimize_from_x(None, dim, spg, wps, elements, calculator,
                                 ref_environments, T, niter, early_quit,
                                 'global', filename=filename,
                                 random_state=random_state)

        if result is not None:
            (xtal, xs) = result
            if xtal.check_validity(criteria, verbose=verbose):
                x = xtal.get_1d_rep_x()
                sim1 = calculate_S(x, xtal, ref_environments, calculator)
                print(xtal.get_xtal_string({'sim': sim1}))
            else:
                xtal = None
                sim1 = None
            return xtal, sim1

        count += 1
        if count == N_max:
            break
    return None, None


def minimize_from_x(x, dim, spg, wps, elements, calculator, ref_environments,
                    T=0.2, niter=20, early_quit=0.02, opt_type='local',
                    minimizers=[('Nelder-Mead', 100), ('L-BFGS-B', 100)],
                    filename='local_opt_data.txt', random_state=None,
                    #derivative=True):
                    derivative=False):
    """
    Generate xtal from the 1d representation

    Args:
        x: list of 1D array
        spg (int): space group number (1-230)
        wps (string): e.g. [['4a', '8b']]
        elements (string): e.g., ['Si', 'O']
    """
    if derivative:
        jac = calculate_dSdx
    else:
        jac = None

    g, wps, dof = get_input_from_letters(spg, wps, dim)
    l_type = g.lattice_type
    sites_wp = []
    sites = []
    numIons = []
    ref_envs = None
    for i, wp in enumerate(wps):
        site = []
        numIon = 0
        for w in wp:
            sites.append((elements[i], w))  # .get_label())
            site.append(w.get_label())
            numIon += w.multiplicity
            if ref_envs is None:
                ref_envs = ref_environments[i]
            else:
                ref_envs = np.vstack((ref_envs, ref_environments[i]))
        sites_wp.append(site)
        numIons.append(numIon)

    if len(ref_envs.shape) == 1:
        ref_envs = ref_envs.reshape((1, len(ref_envs)))

    xtal = pyxtal()
    if x is None:
        count = 0
        while True:
            count += 1
            try:
                xtal.from_random(dim, g, elements, numIons,
                                 sites=sites_wp, factor=1.0,
                                 random_state=random_state)
            except RuntimeError:
                print(g.number, numIons, sites)
                print("Trouble in generating random xtals from pyxtal")
            if xtal.valid:
                atoms = xtal.to_ase(resort=False, add_vaccum=False)
                try:
                    des = calculator.calculate(atoms)['x']
                except:
                    if filename is not None:
                        print('Not a good structure, skip')
                    continue
                x = xtal.get_1d_rep_x()
                break
            elif count == 5:
                return None
    else:
        sites = []
        for ele, _wps in zip(elements, wps):
            for wp in _wps:
                sites.append((ele, wp))
        try:
            xtal.from_1d_rep(x, sites, dim=dim)
        except:
            return None

    x0 = np.array(x.copy())
    # Extract variables, call from Pyxtal
    [N_abc, N_ang] = Lattice.get_dofs(xtal.lattice.ltype)
    rep = xtal.get_1D_representation()
    xyzs = rep.x[1:]

    # Set constraints and minimization
    bounds = [(1.5, 50)] * (N_abc) + [(30, 150)] * (N_ang)

    # Special treatment in case the random lattice is small
    for i in range(N_abc):
        if x[i] < 1.5:
            x[i] = 1.5
        if x[i] > 50.0:
            x[i] = 50.0

    for i in range(N_abc, N_abc + N_ang):
        if x[i] < 30.0:
            x[i] = 30.0
        if x[i] > 150.0:
            x[i] = 150.0

    for xyz in xyzs:
        if len(xyz) > 2:
            bounds += [(0.0, 1.0)] * len(xyz[2:])

    if len(x) != len(bounds):
        print('debug before min', xtal, x, bounds, len(x), len(bounds))

    sim0 = calculate_S(x, xtal, ref_envs, calculator)
    if filename is not None:
        with open(filename, 'a+') as f0:
            f0.write('\nSpace Group: {:d}\n'.format(xtal.group.number))
            for element, numIon, site in zip(elements, numIons, sites_wp):
                strs = 'Element: {:2s} {:4d} '.format(element, numIon)
                for s in site:
                    strs += '{:s} '.format(s)
                strs += '\n'
                f0.write(strs)
            # Initial value
            strs = 'Init: {:9.3f} '.format(sim0)
            for x0 in x:
                strs += '{:8.4f} '.format(x0)
            strs += '\n'
            print(strs)
            f0.write(strs)

    # Run local minimization
    if opt_type == 'local':
        # set call back function for debugging
        def print_local_fun(x):
            f = calculate_S(x, xtal, ref_envs, calculator)
            print("{:.4f} ".format(f), x)
            if filename is not None:
                with open(filename, 'a+') as f0:
                    strs = 'Iter: {:9.3f} '.format(f)
                    for x0 in x[:3]:
                        strs += '{:8.4f} '.format(x0)
                    strs += '\n'
                    f0.write(strs)
        callback = print_local_fun if filename is not None else None

        for minimizer in minimizers:
            (method, step) = minimizer
            #print("Starting", xtal.lattice, method, step)
            if len(x) != len(bounds):
                print('debug min', xtal, x, bounds, len(x), len(bounds))
            res = minimize(calculate_S, x,
                           method=method,
                           args=(xtal, ref_envs, calculator),
                           jac=None if method=='Nelder-Mead' else jac,
                           bounds=bounds,
                           options={'maxiter': step}, #'disp': True},
                           callback=callback)
            x = res.x
            if xtal.lattice is None:
                return None
            #import sys; sys.exit()

        if filename is not None:
            with open(filename, 'a+') as f0:
                f0.write('END\n')
    else:
        # set call back function for debugging
        def print_fun_local(x):
            f = calculate_S(x, xtal, ref_envs, calculator)
            # print("{:.4f} ".format(f), x)
            if f < early_quit:
                return True
            else:
                return None
        callback = print_fun_local  # if verbose else None

        minimizer_kwargs = {'method': ['Nelder-Mead', 'l-bfgs-b',
                                       'Nelder-Mead', 'l-bfgs-b'],
                            'args': (xtal, ref_envs, calculator),
                            'bounds': bounds,
                            'callback': callback,
                            'options': {'maxiter': 100,
                                        'fatol': 1e-6,
                                        'ftol': 1e-6}}

        bounded_step = RandomDispBounds(np.array([b[0] for b in bounds]),
                                        np.array([b[1] for b in bounds]),
                                        id1=N_abc + N_ang,
                                        id2=N_abc)

        # set call back function for debugging
        def print_fun(x, f, accepted):
            if filename is not None:
                print("minimum {:.4f}[{:.4f}] accepted {:d} ".format(
                    f, early_quit, int(accepted)), x[:N_abc])
            if f < early_quit:
                # print("Return True", True is not None)
                return True
            else:
                return None
        callback = print_fun  # if verbose else None

        # Run BH optimization
        res = basinhopping(calculate_S, x, T=T,
                           minimizer_kwargs=minimizer_kwargs,
                           niter=niter,
                           take_step=bounded_step,
                           callback=callback)
        if xtal.lattice is None:
            return None

    # Extract the optimized xtal
    xtal = pyxtal()
    try:
        xtal.from_1d_rep(res.x, sites, dim=dim)
        return xtal, (x0, res.x)
    except:
        return None



class builder(object):
    """
    Class for generating structures with desired local environments

    Args:
        elements (str): e.g. ['Si', 'O']
        composition (int): e.g. [1, 2]
        dim (int): e.g., 0, 1, 2, 3
        db_file (str): default is 'mof.db'
        log_file (str): default is 'mof.log'

    Examples
    --------

    To create a new structure instance

    >>> from pyxtal.lego.builder import builder
    >>> bu = builder(['P', 'O', 'N'], [1, 1, 1], db_file='PON.db')
    """

    def __init__(self, elements, composition, dim=3, prefix='mof',
                 db_file=None, log_file=None, rank=0, verbose=False):

        self.rank = rank
        self.prefix = f"{prefix}-{rank}"
        # Define the chemical system
        self.dim = dim
        self.elements = elements
        self.composition = composition

        # Initialize neccessary functions and attributes
        self.calculator = None       # will be a callable function
        self.ref_environments = None  # will be a numpy array
        self.criteria = {}           # will be a dictionary
        self.verbose = verbose

        # Define the I/O
        logging.getLogger().handlers.clear()
        if log_file is not None:
            self.log_file = log_file
        else:
            self.log_file = self.prefix + '.log'
        logging.basicConfig(format="%(asctime)s| %(message)s",
                            filename=self.log_file,
                            level=logging.INFO)

        self.logging = logging
        if db_file is not None:
            self.db_file = db_file
        else:
            self.db_file = self.prefix + '.db'
        self.db = database_topology(self.db_file, log_file=self.log_file)

    def __str__(self):

        s = "\n------MOF Builder------"
        s += "\nSystem: "
        for element, comp in zip(self.elements, self.composition):
            s += "{:s}{:d} ".format(element, comp)
        s += "\nDatabase: {}".format(self.db_file)
        s += "\nLog_file: {}".format(self.log_file)
        if self.calculator is not None:
            s += "\nDescriptor: {}".format(self.calculator)
        if self.ref_environments is not None:
            (d1, d2) = self.ref_environments.shape
            s += "Reference enviroments ({:}, {:})".format(d1, d2)
        if len(self.criteria.keys()) > 0:
            for key in self.criteria.keys():
                s += '\nCriterion_{:}: {:}'.format(key, self.criteria[key])
        return s

    def __repr__(self):
        return str(self)

    def print_memory_usage(self):
        import psutil
        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss / 1024 ** 2
        self.logging.info(f"Rank {self.rank} memory: {mem:.1f} MB")
        print(f"Rank {self.rank} memory: {mem:.1f} MB")

    def set_descriptor_calculator(self, dtype='SO3', mykwargs={}):
        """
        Set up the calculator for descriptor computation.
        Here we mostly use the pyxtal_ff module

        Arg:
            dytpe (str): only SO3 is suppoted now
        """
        if dtype == 'SO3':
            kwargs = {'lmax': 4,
                      'nmax': 2,
                      'rcut': 2.2,
                      'alpha': 1.5,
                      'weight_on': True,
                      }
            kwargs.update(mykwargs)

            self.calculator = SO3(**kwargs)

    def set_reference_enviroments(self, cif_file, substitute=None):
        """
        Get the reference enviroments

        Args:
            cif_file (str): cif structure
            substitute (dict): substitution directory
        """

        if self.calculator is None:
            raise RuntimeError(
                "Must call set_descriptor_calculator in advance")

        xtal = pyxtal()
        xtal.from_seed(cif_file)
        if substitute is not None:
            xtal.substitute(substitute)  # ; print(xtal)
        xtal.resort_species(self.elements)

        ids = [0] * len(self.elements)
        count = 0
        for site in xtal.atom_sites:
            for i, element in enumerate(self.elements):
                if element == site.specie:
                    ids[i] = count
                    break
            count += site.multiplicity
        if self.verbose:
            print("ids from Reference xtal", ids)
        atoms = xtal.to_ase(resort=False)
        self.ref_environments = self.calculator.compute_p(atoms, ids)
        if self.verbose:
            print(self.ref_environments)
        self.ref_xtal = xtal

    def set_criteria(self, CN=None, dimension=None, min_density=None, exclude_ii=False):
        """
        define the criteria to check if a structure is good

        Args:
            CN (int): coordination number
            dimension (int): target dimensionality (e.g. we want the 3D structure)
            min_density (float): minimum density
            exclude_ii (bool): allow the bond between same element
        """

        if CN is not None:
            self.criteria["CN"] = CN
            if 'cutoff' in CN.keys():
                self.criteria['cutoff'] = CN['cutoff']
            else:
                self.criteria['cutoff'] = None
        if dimension is not None:
            self.criteria["Dimension"] = dimension
        if min_density is not None:
            self.criteria["MIN_Density"] = min_density
        self.criteria['exclude_ii'] = exclude_ii

    def get_input_from_letters(self, spg, wps):
        """
        A short cut to get (spg, wps, dof) from the get_input functions

        Args:
            spg (int): space group number 1-230
            wps (list): e.g. [['4a', '4b']]
        """
        return get_input_from_letters(spg, wps, self.dim)

    def get_input_from_ref_xtal(self, xtal, substitute=None):
        """
        Generate the input from a given pyxtal

        Args:
            xtal: pyxtal object
            substitute: a dictionary to describe chemical substitution
        """

        # Make sure it is a pyxtal object
        if type(xtal) == str:
            c = pyxtal()
            c.from_seed(xtal)
            xtal = c
        if substitute is not None:
            xtal.substitute(substitute)

        g = xtal.group
        sites = [[] for _ in self.elements]
        dof = xtal.lattice.dof
        for s in xtal.atom_sites:
            for i, specie in enumerate(self.elements):
                if s.specie == specie:
                    # wp_combo[i].append(s.wp)
                    sites[i].append(s.wp.get_label())
                    break
            dof += s.wp.get_dof()

        return g.number, sites, dof

    def get_similarity(self, xtal):
        """
        Compute the similrity for a given xtal
        QZ: call calculate_S()

        Args:
            xtal: pyxtal object

        Returns:
            the similarity value .w.r.t the reference environment
        """
        x = xtal.get_1d_rep_x()
        return calculate_S(x, xtal, self.ref_environments,
                           self.calculator)

    def process_xtals(self, xtals, xs, add_db, symmetrize):
        # Now process each of the results
        valid_xtals = []
        count = 0
        for xtal, _xs in zip(xtals, xs):
            status = xtal.check_validity(self.criteria, verbose=self.verbose)
            if status:
                valid_xtals.append(xtal)
                sim1 = self.get_similarity(xtal)
                if symmetrize:
                    pmg = xtal.to_pymatgen()
                    xtal = pyxtal()
                    xtal.from_seed(pmg)
                if add_db:
                    self.process_xtal(xtal, [0, sim1], count, xs=_xs)
                    count += 1
                else:
                    dicts = {'sim': "{:6.3f}".format(sim1)}
                    print(xtal.get_xtal_string(dicts))
        return valid_xtals

    def optimize_xtals(self, xtals, ncpu=1, opt_type='local',
                       T=0.2, niter=20, early_quit=0.02,
                       add_db=True, symmetrize=False,
                       minimizers=[('Nelder-Mead', 100), ('L-BFGS-B', 100)],
                       ):
        """
        Perform optimization for each structure

        Args:
            xtals: list of xtals
            ncpu (int):

        """
        args = (opt_type, T, niter, early_quit, add_db, symmetrize, minimizers)
        if ncpu == 1:
            valid_xtals = self.optimize_xtals_serial(xtals, args)
        else:
            valid_xtals = self.optimize_xtals_mproc(xtals, ncpu, args)
        return valid_xtals

    def optimize_xtals_serial(self, xtals, args):
        """
        Optimization in serial mode.

        Args:
            xtals: list of xtals
            args: (opt_type, T, n_iter, early_quit, add_db, symmetrize, minimizers)
        """
        # (opt_type, T, n_iter, early_quit, add_db, symmetrize, minimizers) = args
        xtals_opt = []
        for i, xtal in enumerate(xtals):
            xtal, sim, _xs = self.optimize_xtal(xtal, i, *args)
            if xtal is not None:
                xtals_opt.append(xtal)
        return xtals_opt

    def optimize_xtals_mproc(self, xtals, ncpu, args):
        """
        Optimization in multiprocess mode.

        Args:
            xtals: list of xtals
            ncpu (int): number of parallel python processes
            args: (opt_type, T, n_iter, early_quit, add_db, symmetrize, minimizers)
        """

        pool = Pool(processes=ncpu)
        (opt_type, T, niter, early_quit, add_db, symmetrize, minimizers) = args
        xtals_opt = deque()

        # Split the input structures to minibatches
        N_batches = 50 * ncpu
        for _i, i in enumerate(range(0, len(xtals), N_batches)):
            start, end = i, min([i+N_batches, len(xtals)])
            ids = list(range(start, end))
            print(f"Rank {self.rank} minibatch {start} {end}")
            self.print_memory_usage()

            def generate_args():
                """
                A generator to yield argument lists for minimize_from_x_par.
                """
                for j in range(ncpu):
                    _ids = ids[j::ncpu]
                    wp_libs = []
                    for id in _ids:
                        xtal = xtals[id]
                        x = xtal.get_1d_rep_x()
                        spg, wps, _ = self.get_input_from_ref_xtal(xtal)
                        wp_libs.append((x, xtal.group.number, wps))
                    yield (self.dim, wp_libs, self.elements, self.calculator,
                           self.ref_environments, opt_type, T, niter,
                           early_quit, minimizers)

            # Use the generator to pass args to reduce memory usage
            _xtal, _xs = None, None
            for result in pool.imap_unordered(minimize_from_x_par,
                                              generate_args(),
                                              chunksize=1):
                if result is not None:
                    (_xtals, _xs) = result
                    valid_xtals = self.process_xtals(
                        _xtals, _xs, add_db, symmetrize)
                    xtals_opt.extend(valid_xtals)  # Use deque to reduce memory

            # Remove the duplicate structures
            self.db.update_row_topology(overwrite=False, prefix=self.prefix)
            self.db.clean_structures_spg_topology(dim=self.dim)

            # After each minibatch, delete the local variables and run garbage collection
            del ids, _xtals, _xs

        xtals_opt = list(xtals_opt)
        print(f"Rank {self.rank} finish optimize_xtals_mproc {len(xtals_opt)}")
        return xtals_opt

    def optimize_reps(self, reps, ncpu=1, opt_type='local',
                      T=0.2, niter=20, early_quit=0.02,
                      add_db=True, symmetrize=False,
                      minimizers=[('Nelder-Mead', 100), ('L-BFGS-B', 100)],
                      N_grids=None):
        """
        Perform optimization for each structure

        Args:
            reps: list of reps
            ncpu (int):
        """
        args = (opt_type, T, niter, early_quit, add_db, symmetrize, minimizers)
        if ncpu == 1:
            valid_xtals = self.optimize_reps_serial(reps, args, N_grids)
        else:
            valid_xtals = self.optimize_reps_mproc(reps, ncpu, args, N_grids)
        return valid_xtals

    def optimize_reps_serial(self, reps, args, N_grids):
        """
        Optimization in multiprocess mode.

        Args:
            reps: list of reps
            ncpu (int): number of parallel python processes
            args: (opt_type, T, n_iter, early_quit, add_db, symmetrize, minimizers)
        """
        xtals_opt = []
        for i, rep in enumerate(reps):
            #print('start', i, rep, len(rep))
            xtal = pyxtal()
            discrete = False if N_grids is None else True
            discrete_cell = abs(rep[1] - round(rep[1])) < 1e-5
            xtal.from_tabular_representation(rep,
                                             normalize=False,
                                             discrete=discrete,
                                             N_grids=N_grids,
                                             discrete_cell=discrete_cell)
                                             #verbose=True)
            xtal, sim, _xs = self.optimize_xtal(xtal, i, *args)
            if xtal is not None:
                xtals_opt.append(xtal)
            #else:
            #    print("Debug===="); import sys; sys.exit()
        return xtals_opt

    def optimize_reps_mproc(self, reps, ncpu, args, N_grids):
        """
        Optimization in multiprocess mode.

        Args:
            reps: list of reps
            ncpu (int): number of parallel python processes
            args: (opt_type, T, n_iter, early_quit, add_db, symmetrize, minimizers)
        """

        pool = Pool(processes=ncpu)
        (opt_type, T, niter, early_quit, add_db, symmetrize, minimizers) = args
        xtals_opt = deque()

        # Split the input structures to minibatches
        N_batches = 50 * ncpu
        for _i, i in enumerate(range(0, len(reps), N_batches)):
            start, end = i, min([i+N_batches, len(reps)])
            ids = list(range(start, end))
            print(f"Rank {self.rank} minibatch {start} {end}")
            self.logging.info(f"Rank {self.rank} minibatch {start} {end}")
            self.print_memory_usage()

            def generate_args():
                """
                A generator to yield argument lists for minimize_from_x_par.
                """
                for j in range(ncpu):
                    _ids = ids[j::ncpu]
                    wp_libs = []
                    for id in _ids:
                        rep = reps[id]
                        xtal = pyxtal()
                        discrete = False if N_grids is None else True
                        discrete_cell = abs(rep[1] - round(rep[1])) < 1e-5
                        xtal.from_tabular_representation(rep,
                                                         normalize=False,
                                                         discrete=discrete,
                                                         discrete_cell=discrete_cell,
                                                         N_grids=N_grids)
                        x = xtal.get_1d_rep_x()
                        spg, wps, _ = self.get_input_from_ref_xtal(xtal)
                        wp_libs.append((x, spg, wps))
                    yield (self.dim, wp_libs, self.elements, self.calculator,
                           self.ref_environments, opt_type, T, niter,
                           early_quit, minimizers)

            # Use the generator to pass args to reduce memory usage
            _xtal, _xs = None, None
            for result in pool.imap_unordered(minimize_from_x_par,
                                              generate_args(),
                                              chunksize=1):
                if result is not None:
                    (_xtals, _xs) = result
                    valid_xtals = self.process_xtals(
                        _xtals, _xs, add_db, symmetrize)
                    xtals_opt.extend(valid_xtals)  # Use deque to reduce memory

            # Remove the duplicate structures
            self.db.update_row_topology(overwrite=False, prefix=self.prefix)
            self.db.clean_structures_spg_topology(dim=self.dim)

            # After each minibatch, delete the local variables and run garbage collection
            del ids, _xtals, _xs
            #gc.collect()  # Explicitly call garbage collector to free memory

        xtals_opt = list(xtals_opt)
        print(f"Rank {self.rank} finish optimize_reps_mproc {len(xtals_opt)}")
        return xtals_opt

    def optimize_xtal(self, xtal, count=0, opt_type='local',
                      T=0.2, niter=20, early_quit=0.02,
                      add_db=True, symmetrize=False,
                      minimizers=[('Nelder-Mead', 100), ('L-BFGS-B', 100)],
                      filename=None):
        """
        Further optimize the input xtal w.r.t reference environment

        Args:
            xtal (instance): pyxtal
        """
        # Change the angle to a better rep
        if xtal.dim == 3 and xtal.lattice is not None and xtal.lattice.ltype in ['triclinic', 'monoclinic']:
            xtal.optimize_lattice(standard=True)
        #xtal.to_file(f'init_{count}.cif')#; print(xtal)
        x = xtal.get_1d_rep_x()
        _, wps, _ = self.get_input_from_ref_xtal(xtal)

        sim0 = self.get_similarity(xtal)
        if xtal.lattice is not None:
            result = minimize_from_x(x, xtal.dim, xtal.group.number, wps,
                                     self.elements, self.calculator,
                                     self.ref_environments,
                                     opt_type=opt_type,
                                     T=T,
                                     niter=niter,
                                     early_quit=early_quit,
                                     minimizers=minimizers,
                                     filename=filename)
            xtal, xs = result
            if result is not None:
                status = xtal.check_validity(self.criteria, verbose=self.verbose)
                sim1 = self.get_similarity(xtal)
                #print("after optim", sim1, status)
            else:
                xtal, xs, status, sim1 = None, None, False, None
        else:
            print("Lattice is None")#, xtal.get_xtal_string())
            xtal, xs, status, sim1 = None, None, False, None
            #import sys; sys.exit()

        if status:
            if symmetrize:
                pmg = xtal.to_pymatgen()
                xtal = pyxtal()
                xtal.from_seed(pmg)
            if add_db:
                self.process_xtal(xtal, [sim0, sim1], count, xs)
            else:
                dicts = {'sim': "{:6.3f} => {:6.3f}".format(sim0, sim1)}
                print(xtal.get_xtal_string(dicts))
        else:
            if self.verbose:
                print('invalid relaxation', count)
                print(xtal.get_xtal_string())
            #import sys; sys.exit()
            #xtal.to_file(f'{count}.cif')
            xtal = None

        return xtal, sim1, xs

    def generate_xtal(self, spg, wps, niter, T=0.2, N_max=5,
                      early_quit=0.03, dump=False, verbose=None,
                      add_db=True, random_state=None):
        """
        Generate a crystal with the desired local environment

        Args:
            spg (int): group number
            wps (list): list of wps for the disired crystal (e.g., [spg, wp1, wp2])
            n_iter (int): number of iterations for basin hopping
            T (float): for basinhopping
            N_max (int): number of maximum
            early_quit (float): threshhold for early termination
            dump (bool): whether or not dump the trajectory
        """
        if verbose is None:
            verbose = self.verbose
        xtal, sim = generate_xtal(self.dim, spg, wps, niter,
                                  self.elements,
                                  self.calculator,
                                  self.ref_environments,
                                  self.criteria,
                                  T, N_max, early_quit, dump,
                                  random_state,
                                  verbose=verbose)

        if xtal is not None and xtal.check_validity(self.criteria):
            if add_db:
                self.process_xtal(xtal, [0, sim], 0)

        return xtal, sim

    def generate_xtals_from_wp_libs(self, wp_libs, N_max=5, ncpu=1,
                                    T=0.2, factor=5, early_quit=0.02):
        """
        Run multiple crystal generation from the given wp_libs
        This is the core part for structure generation.

        Args:
            wp_libs (tuple): (number, spg, wps, dof)
            N_max (int): Number of maximum runs
            ncpu (int): Num of parallel processes
            T (float): basinhopping temperature
            factor (int): the number of Basinhopping iterations = factor * dof
            early_quit (float): early termination for basinhopping
        """

        # Generate xtals
        _args = (self.dim,
                 self.elements,
                 self.calculator,
                 self.ref_environments,
                 self.criteria,
                 T, N_max, early_quit)
        count = 0
        xtals, sims = [], []
        if ncpu == 1:
            for wp_lib in wp_libs:
                (number, spg, wps, dof) = wp_lib
                xtal, sim = generate_xtal(spg, wps, factor*dof, *_args)
                if xtal is not None:
                    xtals.append(xtal)
                    sims.append(sim)

        else:
            N_cycle = int(np.ceil(len(wp_libs)/ncpu))
            print(
                "\n# Parallel Calculation in generate_xtals_from_wp_libs", ncpu, N_cycle)

            args_list = []
            for i in range(ncpu):
                id1 = i * N_cycle
                id2 = min([id1 + N_cycle, len(wp_libs)])
                args_list.append((wp_libs[id1:id2], factor) + _args)

            with ProcessPoolExecutor(max_workers=ncpu) as executor:
                results = [executor.submit(generate_xtal_par, *p)
                           for p in args_list]
                for result in results:
                    (_xtals, _sims) = result.result()
                    xtals.extend(_xtals)
                    sims.extend(_sims)

        for i, xtal in enumerate(xtals):
            self.process_xtal(xtal, [0, sims[i]], i)

        return xtals

    def get_wp_libs_from_spglist(self, spg_list,
                                 num_wp=(None, None),
                                 num_fu=(None, None),
                                 num_dof=(1, 10),
                                 per_spg=30,
                                 ncpu=1):
        """
        Generate wp choices from the list of space groups

        Args:
            spglist (list): list of space group numbers
            num_wp (int): a tuple of (min_wp, max_wp)
            num_fu (int): a tuple of (min_fu, max_fu)
            num_dof (int): a tuple of (min_dof, max_dof)
            per_spg (int): maximum number of wp combinations
            ncpu (int): number of processors
        """

        print('\nGet wp_libs from the given spglist')
        composition = self.composition
        (min_wp, max_wp) = num_wp
        if min_wp is None:
            min_wp = len(composition)
        if max_wp is None:
            max_wp = max([min_wp, len(composition)])
        num_wp = (min_wp, max_wp)

        def process_wp_lib(spg, wp_lib):
            strs = "{:d} wp combos in space group {:d}".format(
                len(wp_lib), spg)
            print(strs)
            self.logging.info(strs)

            if len(wp_lib) > per_spg:
                ids = np.random.choice(
                    range(len(wp_lib)), per_spg, replace=False)
                strs = "Randomly choose {:} wp combinations".format(per_spg)
                wp_lib = [wp_lib[x] for x in ids]
                print(strs)
                self.logging.info(strs)
            return wp_lib

        # Get wp_libs
        wp_libs_total = []
        if ncpu == 1:
            for spg in spg_list:
                wp_lib = generate_wp_lib(
                    [spg], composition, num_wp, num_fu, num_dof)
                wp_lib = process_wp_lib(spg, wp_lib)
                if len(wp_lib) > 0:
                    wp_libs_total.extend(wp_lib)

        else:
            N_cycle = int(np.ceil(len(spg_list)/ncpu))
            args_list = []
            print("\n# Parallel Calculation in get_wp_libs_from_spglist", ncpu, N_cycle)
            for i in range(ncpu):
                id1 = i*N_cycle
                id2 = min([id1+N_cycle, len(spg_list)])
                args_list.append((spg_list[id1:id2],
                                 composition,
                                 num_wp,
                                 num_fu,
                                 num_dof))

            # collect the results
            with ProcessPoolExecutor(max_workers=ncpu) as executor:
                results = [executor.submit(generate_wp_lib_par, *p)
                           for p in args_list]
                for result in results:
                    (spgs, wp_libs) = result.result()
                    for spg, wp_lib in zip(spgs, wp_libs[0]):
                        wp_lib = process_wp_lib(spg, wp_lib)
                        wp_libs_total.extend(wp_lib)

        return sorted(wp_libs_total)

    def get_wp_libs_from_xtals(self, db_file=None,
                               num_wp=(None, None),
                               num_dof=(1, 20),
                               num_atoms=[1, 500]):
        """
        For each struc in the database, get the (spg, wp) info

        Args:
            df_file (str): database file
            num_wp (int): tuple of (min_wp, max_wp)
            num_dof (int): tuple of (min_dof, max_dof)
            num_atoms (int): tuple of (min_at, max_at)
        """

        (min_dof, max_dof) = num_dof
        (min_wp, max_wp) = num_wp
        (min_at, max_at) = num_atoms
        if min_wp is None:
            min_wp = len(self.composition)
        if max_wp is None:
            max_wp = max([min_wp, len(self.composition)])

        wp_libs_total = []
        if db_file is not None:
            strs = "====Loading the structures from {:}".format(db_file)
            print("\n", strs)
            self.logging.info(strs)

            with connect(db_file) as db:
                for row in db.select():
                    atoms = db.get_atoms(row.id)
                    if min_at <= len(atoms) <= max_at:
                        xtal = pyxtal()
                        xtal.from_seed(atoms, tol=0.1)
                        if min_wp <= len(xtal.atom_sites) <= max_wp and \
                                min_dof <= xtal.get_dof() <= max_dof and \
                                xtal.check_validity(self.criteria):
                            t0 = time()
                            sim0 = self.get_similarity(xtal)
                            dicts = {'sim': "0 => {:6.3f}".format(sim0)}
                            strs = xtal.get_xtal_string(dicts)
                            print(strs, "{:.6f}".format(time()-t0))
                            spg, wps, dof = self.get_input_from_ref_xtal(xtal)
                            wp_libs_total.append(
                                (sum(xtal.numIons), spg, wps, dof))

        return sorted(wp_libs_total)

    def import_structures(self, db_file, ids=(None, None),
                          check=True, same_group=True,
                          spglist=range(1, 231), bounds=[1, 500],
                          relax=True,
                          ):
        """
        Import the structures from the external ase database

        Args:
            db_file (str): ase database
            check (boolean): whether or not
            spglist (int): list of spg numbers
            bounds: number of atoms
            energy (boolean): add energy or not
        """
        [lb, ub] = bounds
        count = 0
        strs = "====Loading the structures from {:}".format(db_file)
        print("\n", strs)
        self.logging.info(strs)
        with connect(db_file) as db:
            (min_id, max_id) = ids
            if min_id is None:
                min_id = 1
            if max_id is None:
                max_id = db.count() + 100000
            for row in db.select():
                if min_id <= row.id <= max_id:
                    atoms = row.toatoms()
                    xtal = pyxtal()
                    try:
                        xtal.from_seed(atoms, tol=0.1)
                        if lb <= len(atoms) <= ub and xtal.group.number in spglist:
                            status = True
                        else:
                            status = False
                    except:
                        print("Error in loading structure")
                        status = False

                    # Check structures
                    if status:
                        # Relax the structure
                        if relax:
                            xtal, sim, _ = self.optimize_xtal(
                                xtal, add_db=False)

                        if xtal is not None:
                            energy = row.ff_energy if hasattr(
                                row, 'ff_energy') else None
                            topology = row.topology if hasattr(
                                row, 'topology') else None
                            self.process_xtal(xtal, [0, sim], count,
                                              energy=energy,
                                              topology=topology,
                                              same_group=same_group)
                            count += 1

    def process_xtal(self, xtal, sim, count=0, xs=None, energy=None,
                     topology=None, same_group=True, db=None, check=False):
        """
        Check, print and add xtal to the database

        Args:
            xtal: pyxtal object
            sim: list of two similarity numbers
            count (int): id of the structure in the database
            xs (tuple): list of reps before and after optimization
            energy (float): the energy value
            same_group (bool): keep the same group or not
            db (str): db path
            check (bool): whether or not check if the structure is a duplicate
        """
        if db is None:
            db = self.db
        if check:
            status = db.check_new_structure(xtal, same_group)
        else:
            status = True
        header = "{:4d}".format(count)
        dicts = {'energy': energy,
                 'status': status,
                 'sim': "{:12.3f} => {:6.3f}".format(sim[0], sim[1])
                 }
        strs = xtal.get_xtal_string(dicts, header)
        print(strs)
        self.logging.info(strs)
        kvp = {
            'similarity0': sim[0],
            'similarity': sim[1],
        }
        if xs is not None:
            kvp['x_init'] = np.array2string(xs[0])
            kvp['x_opt'] = np.array2string(xs[1])
        if energy is not None:
            kvp['ff_energy'] = energy
        if topology is not None:
            kvp['topology'] = topology
        if status:
            db.add_xtal(xtal, kvp)


"""
Custom step-function for basin hopping optimization
"""


class RandomDispBounds(object):
    """
    random displacement with bounds:
    see: https://stackoverflow.com/a/21967888/2320035
    """

    def __init__(self, xmin, xmax, id1, id2, stepsize=0.1, dumpfile=None):
        self.xmin = xmin  # bound_min
        self.xmax = xmax  # bound_max
        self.id1 = id1
        self.id2 = id2
        self.stepsize = stepsize
        self.dumpfile = dumpfile

    def __call__(self, x):
        """
        Move the step proportionally within the bound
        """
        # To strongly rebounce the values hitting the wall
        for i in range(self.id2):
            if abs(x[i] - self.xmax[i]) < 0.1:
                # print("To strongly rebounce max values", x[i], self.xmax[i])
                x[i] *= 0.5
            elif abs(x[i] - self.xmin[i]) < 0.1:
                # print("To strongly rebounce min values", x[i], self.xmin[i])
                x[i] *= 2.0

        random_step = np.random.uniform(low=-self.stepsize,
                                        high=self.stepsize,
                                        size=x.shape)
        xnew = x + random_step
        # Cell
        coefs = 1+0.2*(np.random.sample(self.id2)-0.5)
        xnew[:self.id2] *= coefs

        # xyz
        xnew[self.id1:] -= np.floor(xnew[self.id1:])
        # xnew = np.maximum(self.xmin, xnew)
        # xnew = np.minimum(self.xmax, xnew)

        # Randomly introduce compression to prevent non-pd xtal
        if np.random.random() < 0.5:
            id = np.argmax(xnew[:self.id2])
            xnew[id] *= 0.8
        xnew = np.maximum(self.xmin, xnew)
        xnew = np.minimum(self.xmax, xnew)

        # min_step = np.maximum(self.xmin - x, -self.stepsize)
        # max_step = np.minimum(self.xmax - x, self.stepsize)
        # random_step = np.random.uniform(low=min_step, high=max_step, size=x.shape)
        # xnew = x + random_step
        if self.dumpfile is not None:
            with open(self.dumpfile, 'a+') as f0:
                # Initial value
                strs = 'Init: {:9.3f} '.format(10.0)
                for x0 in xnew:
                    strs += '{:8.4f} '.format(x0)
                strs += '\n'
                f0.write(strs)

        return xnew


if __name__ == "__main__":

    xtal = pyxtal()
    xtal.from_spg_wps_rep(194, ['2c', '2b'], [2.46, 6.70])
    cif_file = xtal.to_pymatgen()
    bu = builder(['C'], [1], db_file='reaxff.db',
                          verbose=False)  # True)
    bu.set_descriptor_calculator(mykwargs={'rcut': 1.9})
    bu.set_reference_enviroments(cif_file)
    bu.set_criteria(CN={'C': [3]})
    print(bu)
    print(bu.ref_xtal)
    if False:
        wp_libs = bu.get_wp_libs_from_spglist([191, 179], ncpu=1)
        for wp_lib in wp_libs[:4]:
            print(wp_lib)
        bu.generate_xtals_from_wp_libs(wp_libs[4:8], ncpu=2, N_max=4, early_quit=0.05)

    if False:
        spg, wps = 179, ['6a', '6a', '6a', '6a']
        xtals = []
        for x in [
            # [ 9.6244, 2.5459, 0.1749, 0.7701, 0.4501, 0.6114],
            # [15.0223, 1.5013, 0.8951, 0.6298, 0.4530, 0.1876],
            # [10.0129, 2.6424, 0.3331, 0.7246, 0.4719, 0.8628],
            # [10.2520, 3.1457, 0.2367, 0.6994, 0.2522, 0.6533],
            # [ 9.3994,   2.5525,   0.3072,   0.8414,   0.3480,   0.9638],
            [11.1120,   2.6428,   0.2973,   0.7513,   0.4236,   0.8777],
            [7.9522,   2.6057,   0.5922,   0.9268,   0.6081,   0.3077],
        ]:
            xtal = pyxtal()
            xtal.from_spg_wps_rep(spg, wps, x, ['C']*len(wps))
            xtals.append(xtal)
            bu.optimize_xtal(xtal)
        bu.optimize_xtals(xtals)



import numpy as np

from pyxtal.lattice import Lattice
from pyxtal.molecule import find_rotor_from_smile
from pyxtal.symmetry import Group
from pyxtal.wyckoff_site import atom_site, mol_site

class representation_atom:
    """
    A class to handle the 1D representation of atomic crystal
    Works for Zprime > 1

    Args:
        x: a list of [cell, site_1, site_2, ...]
    """

    def __init__(self, x):
        self.x = x

    def __str__(self):
        return self.to_string()

    @classmethod
    def from_pyxtal(cls, struc, standard=True):
        """
        Initialize 1D rep. from the pyxtal object

        Args:
            struc: pyxtal object
        """
        if standard and not struc.standard_setting:
            pmg = struc.to_pymatgen()
            struc.from_seed(pmg, standard=True)
        symmetry = [struc.atom_sites[0].wp.hall_number]
        lat = struc.lattice.encode()
        vector = [symmetry + lat]
        vector.extend([site.encode() for site in struc.atom_sites])
        x = vector
        return cls(x)

    def to_standard_setting(self):
        xtal = self.to_pyxtal()
        self.x = representation.from_pyxtal(xtal, standard=True).x

    def to_pyxtal(self):
        """
        Export the pyxtal structure

        Args:
            smiles: list of smiles
            compoisition: list of composition
        """
        from pyxtal import pyxtal

        # symmetry
        v = self.x[0]
        struc = pyxtal()
        struc.group, _number = Group(v[0], use_hall=True), v[0]

        # lattice
        ltype = struc.group.lattice_type
        if ltype == "triclinic":
            a, b, c, alpha, beta, gamma = v[1], v[2], v[3], v[4], v[5], v[6]
        elif ltype == "monoclinic":
            a, b, c, alpha, beta, gamma = v[1], v[2], v[3], 90, v[4], 90
        elif ltype == "orthorhombic":
            a, b, c, alpha, beta, gamma = v[1], v[2], v[3], 90, 90, 90
        elif ltype == "tetragonal":
            a, b, c, alpha, beta, gamma = v[1], v[1], v[2], 90, 90, 90
        elif ltype == "hexagonal":
            a, b, c, alpha, beta, gamma = v[1], v[1], v[2], 90, 90, 120
        else:
            a, b, c, alpha, beta, gamma = v[1], v[1], v[1], 90, 90, 90
        try:
            struc.lattice = Lattice.from_para(a, b, c, alpha, beta, gamma, ltype=ltype)
        except:
            print(a, b, c, alpha, beta, gamma, ltype)
            raise ValueError("Problem in Lattice")

        # sites
        struc.numIons = []
        struc.atom_sites = []
        species = []

        for _x in self.x[1:]:
            dicts = {}
            specie, index, pos = _x[0], _x[1], _x[2:]
            dicts["specie"] = specie
            dicts["index"] = index
            dicts["dim"] = 3
            dicts["PBC"] = [1, 1, 1]
            dicts["hn"] = struc.group.hall_number
            wp = struc.group[index]
            dicts["position"] = wp.get_position_from_free_xyzs(pos)
            site = atom_site.load_dict(dicts)
            struc.atom_sites.append(site)

            if specie not in species:
                species.append(specie)
                struc.numIons.append(wp.multiplicity)
            else:
                for i, _specie in enumerate(species):
                    if _specie == specie:
                        struc.numIons[i] += site.wp.multiplicity

        struc.species = species
        struc._get_formula()
        struc.source = "1D rep."
        struc.valid = True
        struc.standard_setting = site.wp.is_standard_setting()

        return struc

    def to_array(self):
        """
        Export only varibles to a 1d numpy array
        """
        cells, xyzs = self.x[0][1:], self.x[1:]
        x = cells
        for xyz in xyzs:
            x = np.hstack((x, xyz[2:]))
        return x

    def to_string(self, time=None, eng=None, tag=None):
        """
        Export string representation

        Args:
            time: float
            eng: float
            tag: string
        """
        x = self.x
        strs = f"{int(x[0][0]):3d} "

        # data for cell
        if x[0][0] <= 348:
            num = 4
        elif x[0][0] <= 488:
            num = 3
        else:  # cubic
            num = 2

        for c in x[0][1:num]:
            strs += f"{c:5.2f} "
        for c in x[0][num:]:
            strs += f"{c:5.1f} "

        # data for atoms
        strs += f"{len(x) - 1:d} "  # Number of sites
        for i in range(1, len(x)):
            strs += f"{x[i][0]:s} "
            strs += f"{x[i][1]:d} "
            for v in x[i][2:]:
                strs += f"{v:6.4f} "

        if time is not None:
            strs += f"{time:5.2f}"

        if eng is not None:
            strs += f"{eng:11.3f}"

        if tag is not None:
            strs += f" {tag:s}"

        return strs


class representation:
    """
    A class to handle the 1D representation of molecular crystal
    Works for Zprime > 1

    Args:
        x: a list of [cell, site_1, site_2, ...]
        smiles: a list of [smile_1, smile_2, ...]
    """

    def __init__(self, x, smiles=None):
        if smiles is not None:
            self.smiles = []
            for _i, smile in enumerate(smiles):
                if smile.endswith(".smi"):
                    smile = smile[:-4]
                self.smiles.append(smile)
        else:
            self.smiles = None
        self.x = x

    def __str__(self):
        return self.to_string()

    @classmethod
    def from_pyxtal(cls, struc, standard=False):
        """
        Initialize 1D rep. from the pyxtal object

        Args:
            struc: pyxtal object
        """
        if standard and not struc.standard_setting:
            # struc.optimize_lattice(standard=True)
            pmg = struc.to_pymatgen()
            struc.from_seed(pmg, molecules=struc.molecules, standard=True)
        symmetry = [struc.mol_sites[0].wp.hall_number]
        lat = struc.lattice.encode()
        vector = [symmetry + lat]
        smiles = []
        for site in struc.mol_sites:
            vector.append(site.encode())
            smiles.append(site.molecule.smile)
        x = vector
        if smiles[0] is None: smiles = None
        return cls(x, smiles)

    @classmethod
    def from_string(cls, inputs, smiles, composition=None):
        """
        Initialize 1D rep. from the string

        Args:
            inputs: input string
            smiles: list of smiles
        """
        # parse the cell
        if composition is None:
            composition = [1] * len(smiles)

        inputs = [float(tmp) for tmp in inputs.split()]
        hn = int(inputs[0])
        if hn <= 2:
            n_cell = 8
        elif hn <= 107:
            n_cell = 6
        elif hn <= 348:
            n_cell = 5
        elif hn <= 488:
            n_cell = 4
        else:
            n_cell = 3  # cubic
        cell = [hn] + inputs[1 : n_cell - 1]

        x = [cell]
        n_site = int(inputs[n_cell - 1])
        if n_site != sum(composition):
            msg = f"Composition is inconsistent: {sum(composition):d}/{n_site:d}\n"
            msg += str(inputs)
            raise ValueError(msg)
        # n_cell += 1

        for i, smile in enumerate(smiles):
            if smile.endswith(".smi"):
                smile = smile[:-4]
            for _c in range(composition[i]):
                if smile in ["Cl-"]:
                    n_mol = 4
                else:
                    n_torsion = len(find_rotor_from_smile(smile))
                    n_mol = 8 + n_torsion  # (wp_id, x, y, z, ori_x, ori_y, ori_z, inv) + torsion
                # inversion
                # print(n_mol, n_cell, len(inputs))
                inputs[n_cell] = int(inputs[n_cell])
                inputs[n_cell + n_mol - 1] = int(inputs[n_cell + n_mol - 1])
                x.append(inputs[n_cell : n_cell + n_mol])  # ; print('string', x[-1])
                n_cell += n_mol
        assert n_cell == len(inputs)
        return cls(x, smiles)

    def to_standard_setting(self):
        xtal = self.to_pyxtal()
        xtal.optimize_lattice(standard=True)
        rep0 = representation.from_pyxtal(xtal)
        self.x = rep0.x

    def to_pyxtal(self, smiles=None, composition=None):
        """
        Export the pyxtal structure

        Args:
            smiles: list of smiles
            compoisition: list of composition
        """
        from pyxtal import pyxtal

        if smiles is None:
            smiles = self.smiles

        if composition is None:
            composition = [1] * len(smiles)

        if sum(composition) + 1 != len(self.x):
            msg = "Composition is inconsistent:\n"
            msg += str(composition) + "\n"
            msg += self.to_string()
            raise ValueError(msg)

        # symmetry
        v = self.x[0]
        struc = pyxtal(molecular=True)
        struc.group, _number = Group(v[0], use_hall=True), v[0]

        # lattice
        ltype = struc.group.lattice_type
        if ltype == "triclinic":
            a, b, c, alpha, beta, gamma = v[1], v[2], v[3], v[4], v[5], v[6]
        elif ltype == "monoclinic":
            a, b, c, alpha, beta, gamma = v[1], v[2], v[3], 90, v[4], 90
        elif ltype == "orthorhombic":
            a, b, c, alpha, beta, gamma = v[1], v[2], v[3], 90, 90, 90
        elif ltype == "tetragonal":
            a, b, c, alpha, beta, gamma = v[1], v[1], v[2], 90, 90, 90
        elif ltype == "hexagonal":
            a, b, c, alpha, beta, gamma = v[1], v[1], v[2], 90, 90, 120
        else:
            a, b, c, alpha, beta, gamma = v[1], v[1], v[1], 90, 90, 90
        try:
            struc.lattice = Lattice.from_para(a, b, c, alpha, beta, gamma, ltype=ltype)
        except:
            print(a, b, c, alpha, beta, gamma, ltype)
            raise ValueError("Problem in Lattice")

        # sites
        struc.numMols = [0] * len(smiles)
        struc.molecules = []
        struc.mol_sites = []

        count = 1
        for i, comp in enumerate(composition):
            smile = smiles[i]
            if smile.endswith(".smi"):
                smile = smile[:-4]
            for _j in range(comp):
                v = self.x[count]
                dicts = {}
                dicts["smile"] = smile
                dicts["type"] = i
                dicts["dim"] = 3
                dicts["PBC"] = [1, 1, 1]
                # dicts['number'] = number
                dicts["hn"] = struc.group.hall_number
                dicts["index"] = v[0]
                dicts["lattice"] = struc.lattice.matrix
                dicts["lattice_type"] = ltype
                dicts["center"] = v[1:4]
                if smile not in ["Cl-"]:
                    dicts["orientation"] = np.array(v[4:7])
                    dicts["rotor"] = v[7:-1]  # ; print('ro', dicts['rotor'])
                    dicts["reflect"] = int(v[-1])
                site = mol_site.from_1D_dicts(dicts)

                bypass = False
                for mol_id, molecule in enumerate(struc.molecules):
                    if str(site.molecule) == str(molecule):
                        site.type = mol_id
                        struc.numMols[mol_id] += site.wp.multiplicity
                        bypass = True
                        break
                if not bypass:
                    struc.molecules.append(site.molecule)
                    site.type = len(struc.molecules) - 1
                    struc.numMols[site.type] += site.wp.multiplicity

                # site.type = i
                struc.mol_sites.append(site)
                # move to next rep
                count += 1

        struc._get_formula()
        struc.source = "1D rep."
        struc.valid = True
        struc.standard_setting = site.wp.is_standard_setting()

        return struc

    def to_string(self, time=None, eng=None, tag=None):
        """
        Export string representation

        Args:
            time: float
            eng: float
            tag: string
        """
        x = self.x
        strs = f"{int(x[0][0]):3d} "

        # data for cell
        if x[0][0] <= 348:
            num = 4
        elif x[0][0] <= 488:
            num = 3
        else:  # cubic
            num = 2

        for c in x[0][1:num]:
            strs += f"{c:5.2f} "
        for c in x[0][num:]:
            strs += f"{c:5.1f} "

        # data for molecule
        strs += f"{len(x) - 1:d} "  # ; print(x[1])
        for i in range(1, len(x)):
            strs += f"{x[i][0]:d} "
            for v in x[i][1:4]:
                strs += f"{v:4.2f} "
            for v in x[i][4:-1]:
                strs += f"{v:6.1f} "
            strs += f"{int(x[i][-1]):d} "

        if time is not None:
            strs += f"{time:5.2f}"

        if eng is not None:
            strs += f"{eng:11.3f}"

        if tag is not None:
            strs += f" {tag:s}"

        return strs

    def same_smiles(self, smiles):
        if len(self.smiles) == smiles:
            return all(s2 == s2 for s1, s2 in zip(self.smiles, smiles))
        else:
            return False

    def get_dist(self, rep):
        """
        get distance with the other rep1
        Now only supports Z'=1
        """
        from pyxtal.symmetry import Wyckoff_position as WP

        if self.same_smiles(rep.smiles):
            msg = "different smiles"
            print(msg)
            return None
        elif len(self.x) != len(rep.x):
            msg = "different number of sites"
            print(msg)
            return None
        elif self.x[0][0] != rep.x[0][0]:
            msg = "different space group numbers"
            print(msg)
            return None
        else:
            diffs = []
            wp = WP.from_group_and_index(self.x[0][0], 0, use_hall=True)
            for i in range(len(self.x)):
                np.zeros(len(self.x[i]))
                tmp1 = np.array(self.x[i])
                tmp2 = np.array(rep.x[i])
                # cell difference
                if i == 0:
                    diff_cell = tmp2 - tmp1
                    diffs.extend(diff_cell)
                # site difference
                else:
                    # symmmetry variation
                    xyzs = wp.apply_ops(tmp2[:3])
                    diff_xyzs = xyzs - tmp1[:3]
                    diff_xyzs -= np.rint(diff_xyzs)
                    id = np.argmin(np.linalg.norm(diff_xyzs, axis=1))
                    diff_xyz = diff_xyzs[id]
                    diff_ori = tmp2[3:6] - tmp1[3:6]
                    diff_ori /= [360.0, 180.0, 360.0]
                    diff_ori -= np.rint(diff_ori)
                    diff_ori *= [360.0, 180.0, 360.0]
                    diff_tor = tmp2[6:] - tmp1[6:]
                    diff_tor /= 360.0
                    diff_tor -= np.rint(diff_tor)
                    diff_tor *= 360.0
                    diffs.extend(diff_xyz)
                    diffs.extend(diff_ori)
                    diffs.extend(diff_tor)
            return np.array(diffs)


if __name__ == "__main__":
    # aspirin
    smiles = ["CC(=O)OC1=CC=CC=C1C(=O)O"]
    x = [
        [81, 11.43, 6.49, 11.19, 83.31],
        [0, 0.77, 0.57, 0.53, 48.55, 24.31, 145.94, -77.85, -4.40, 170.86, False],
    ]
    # rep0 = representation(x, smiles)
    # print(rep0.to_string())
    rep1 = representation(x, smiles)
    xtal = rep1.to_pyxtal()
    print(xtal)
    rep2 = representation.from_pyxtal(xtal)
    print(rep2.to_pyxtal())
    print(rep2.to_string())

    print("Test read from string")
    string = "82 11.43  6.49 11.19 83.31 1  0 0.77  0.57  0.53 48.55 24.31 145.9 -77.85 -4.40 170.9 0"
    rep3 = representation.from_string(string, smiles)
    print(rep3.to_string())
    print(rep3.to_pyxtal())
    # x = rep3.to_pyxtal(); x.optimize_lattice(standard=True); print(x)
    rep3.to_standard_setting()
    print(rep3.to_pyxtal())
    print(rep3.to_string())

    print("Test other cases")
    string1 = "81 14.08  6.36 25.31  83.9 1 0 0.83 0.40 0.63  136.6  -21.6 -151.1 -101.1 -131.2  154.7 -176.4 -147.8  178.2 -179.1  -53.3 0"  # noqa: E501
    string2 = "81 14.08  6.36 25.31  83.9 1 0 0.03 0.84 0.89  149.1   -8.0  -37.8  -39.9 -104.2  176.2 -179.6  137.8 -178.5 -173.3 -103.6 0"  # noqa: E501
    smiles = ["CC1=CC=C(C=C1)S(=O)(=O)C2=C(N=C(S2)C3=CC=C(C=C3)NC(=O)OCC4=CC=CC=C4)C"]
    rep4 = representation.from_string(string1, smiles)
    rep5 = representation.from_string(string2, smiles)
    print(string1)
    print(string2)
    print(rep4.get_dist(rep5))

    from pyxtal import pyxtal

    xtal = pyxtal()
    xtal.from_seed("pyxtal/database/cifs/Fd3.cif")
    xtal.from_seed("pyxtal/database/cifs/NaSb3F10.cif")
    rep = representation_atom.from_pyxtal(xtal)
    print(rep)
    print(xtal)
    print(rep.to_pyxtal())
    # strings = [
    # "83 14.08 6.36 25.31 83.9 1 0.72 0.40 0.27  131.6 -17.0 -120.0  -83.8 -134.1 -174.5 -175.7 -168.8  173.9  178.0 -157.4 0",
    # "81 14.08 6.36 25.31 83.9 1 0.59 0.81 0.39 -117.8 -50.1  -95.3  -25.8  -80.6  164.7  155.9 -124.9 -159.2  178.6 -154.7 0",
    # "81 14.08 6.36 25.31 83.9 1 0.75 0.09 0.01  133.8 -19.5  -55.1  -86.7  -91.7 -175.0 -170.4 -176.8  173.3 -164.8  -58.4 0",
    # "81 14.08 6.36 25.31 83.9 1 0.72 0.44 0.01  135.2  27.5   97.2 -101.1 -105.1  -29.7 -169.7  -50.1  172.2 -173.1  131.6 0",
    # "82 14.00 6.34 25.26 83.6 1 0.21 0.08 0.54  146.0 -12.0   50.2  108.0  112.3 -166.3 -158.7  -35.5  172.3 -168.7  133.0 0",
    # "81 14.08 6.36 25.31 83.9 1 0.05 0.30 0.89  -68.2  41.2  148.8  -66.9  -85.0 -167.4  172.3 -166.2 -178.3  166.4  -45.9 0",
    # ]

    # import pymatgen.analysis.structure_matcher as sm
    # matcher = sm.StructureMatcher(ltol=0.3, stol=0.3, angle_tol=10)
    # for i, string in enumerate(strings):
    #    print(str(i) + '  ' +string)
    #    rep4 = representation.from_string(string, smiles)
    #    pmg1 = rep4.to_pyxtal().to_pymatgen(); pmg1.remove_species('H')
    #    rep4.to_standard_setting()
    #    pmg2 = rep4.to_pyxtal().to_pymatgen(); pmg2.remove_species('H')
    #    print(i, rep4.to_string(), matcher.fit(pmg1, pmg2))



import numpy as np
from pyxtal.symmetry import Group
from pyxtal import pyxtal

def get_input_from_letters(spg, sites, dim=3):
    """
    Prepare the input from letters

    Args:
        spg (int): 1-230
        sites (list): [['4a'], ['4b']] or [[0], [1]]
        dim (int): 0, 1, 2, 3

    Returns:
        space group instance + wp_list + dof
    """
    g = Group(spg, dim=dim)
    wp_combo = []
    dof = g.get_lattice_dof()
    for site in sites:
        wp_specie = []
        for s in site:
            if type(s) in [int, np.int64]:
                wp = g[s]
            else:
                wp = g.get_wyckoff_position(s)
            wp_specie.append(wp)
            dof += wp.get_dof()
        wp_combo.append(wp_specie)
    return g, wp_combo, dof

def create_trajectory(dumpfile, trjfile, modes=['Init', 'Iter'], dim=3):
    """
    create the ase trajectory from the dump file.
    This is used to analyze the performance of structure optimization
    """

    from ase.io import write

    keyword_start = 'Space Group'  # \n'
    keyword_end = 'END'  # \n'

    with open(dumpfile, 'r') as f:
        lines = f.readlines()
        starts, ends = [], []
        for i in range(len(lines)):
            l = lines[i].lstrip()
            if l.startswith(keyword_start):
                starts.append(i)
            elif l.startswith(keyword_end):
                ends.append(i)
    # print(len(starts), len(ends))
    assert (len(starts) == len(ends))

    atoms = []
    for i, j in zip(starts, ends):
        # parse each run
        xtal_strs = lines[i:j]
        spg = int(xtal_strs[0].split(':')[1])
        l_type = Group(spg).lattice_type
        elements = []
        numIons = []
        wps = []
        for count, tmp in enumerate(xtal_strs[1:]):
            if tmp.startswith('Element'):
                tmp1 = tmp.split(':')
                tmp2 = tmp1[1].split()
                elements.append(tmp2[0])
                numIons.append(int(tmp2[1]))
                wps.append(tmp2[2:])
            else:
                line_struc = count + 1
                break

        # print(spg, elements, numIons, wps)
        g, wps, dof = get_input_from_letters(spg, wps, dim)
        for line_number in range(line_struc, len(xtal_strs)):
            tmp0 = xtal_strs[line_number].split(':')
            tag = tmp0[0]
            if tag in modes:
                tmp = tmp0[1].split()
                sim = float(tmp[0])
                x = [float(t) for t in tmp[1:]]

                # QZ: to fix
                xtal = pyxtal()
                xtal.from_1d_rep(x, wps, numIons, l_type, elements, dim)

                struc = xtal.to_ase()
                struc.info = {'time': line_number-line_struc, 'fun': sim}
                atoms.append(struc)
                # print(xtal.lattice)
        write(filename=trjfile, images=atoms, format='extxyz')


def calculate_dSdx(x, xtal, des_ref, f, eps=1e-4, symmetry=True, verbose=False):
    """
    Compute dSdx via the chain rule of dSdr*drdx.
    This routine will serve as the jacobian for scipy.minimize

    Args:
        x (float): input variable array to describe a crystal structure
        xtal (instance): pyxtal object
        des_ref (array): reference environment
        f (callable): function to compute the environment
        symmetry (bool): whether or not turn on symmetry
        verbose (bool): output more information
    """
    xtal.update_from_1d_rep(x)
    ids = [0]
    weights = []
    for site in xtal.atom_sites:
        ids.append(site.wp.multiplicity + ids[-1])
        weights.append(site.wp.multiplicity)
    ids = ids[:-1]
    weights = np.array(weights, dtype=float)
    atoms = xtal.to_ase()
    dPdr, P = f.compute_dpdr_5d(atoms, ids)
    dPdr = np.einsum('i, ijklm -> ijklm', weights, dPdr)

    # Compute dSdr [N, M] [N, N, M, 3, 27] => [N, 3, 27]
    # print(P.shape, des_ref.shape, dPdr.shape)
    dSdr = np.einsum("ik, ijklm -> jlm", 2*(P - des_ref), dPdr)

    # Get supercell positions
    ref_pos = np.repeat(atoms.positions[:, :, np.newaxis], 27, axis=2)
    cell_shifts = np.dot(VECTORS, atoms.cell)
    ref_pos += cell_shifts.T[np.newaxis, :, :]

    # Compute drdx via numerical func
    drdx = np.zeros([len(atoms), 3, 27, len(x)])

    xtal0 = xtal.copy()
    for i in range(len(x)):
        x0 = x.copy()
        x0[i] += eps
        # Maybe expensive Reduce calls, just need positions
        xtal0.update_from_1d_rep(x0)
        atoms = xtal0.to_ase()

        # Get supercell positions
        pos = np.repeat(atoms.positions[:, :, np.newaxis], 27, axis=2)
        cell_shifts = np.dot(VECTORS, atoms.cell)
        pos += cell_shifts.T[np.newaxis, :, :]
        drdx[:, :, :, i] += (pos - ref_pos)/eps

    # [N, 3, 27] [N, 3, 27, H] => H
    dSdx = np.einsum("ijk, ijkl -> l", dSdr, drdx)
    return dSdx

def calculate_S(x, xtal, des_ref, f, return_rdf=False):
    """
    Optimization function used for structure generation

    Args:
        x (float): input variable array to describe a crystal structure
        xtal (instance): pyxtal object
        des_ref: reference environment
        f (callable): function to compute the enivronment
        return_rdf (bool): whether to return the radial distribution function

    Returns:
        obj (float): objective function value
    """
    if xtal.lattice is None:
        des = np.zeros(des_ref.shape)
        weights = 1
    else:
        xtal.update_from_1d_rep(x)  # ; print(xtal)
        if xtal.lattice is not None:
            ids = [0]
            weights = []
            for site in xtal.atom_sites:
                ids.append(site.wp.multiplicity + ids[-1])
                weights.append(site.wp.multiplicity)
            ids = ids[:-1]
            weights = np.array(weights, dtype=float)
            atoms = xtal.to_ase(resort=False)
            #des = f.calculate(atoms)['x'][ids]
            des = f.compute_p(atoms, ids, return_rdf=return_rdf)
        else:
            des = np.zeros(des_ref.shape)
            weights = 1

    sim = np.sum((des-des_ref)**2, axis=1)  # /des.shape[1]
    #if ref_CN is not None:
    #    coefs = CNs[:, 1] - ref_CN
    #    coefs[coefs < 0] = 0
    #    penalty = coefs * (f.rcut + 0.01 - CNs[:, 0])
    #    if penalty.sum() > 0: print('debug', penalty.sum(), x[:3])
    #    sim += 5 * coefs * penalty
    obj = np.sum(sim * weights) / sum(xtal.numIons)
    #print(des-des_ref); print(sim); print(obj)#; import sys; sys.exit()
    return obj



"""
Module for generating molecular crystals
"""

# Standard Libraries
from copy import deepcopy
import numpy as np

from pyxtal.lattice import Lattice
from pyxtal.molecule import pyxtal_molecule
from pyxtal.msg import Comp_CompatibilityError, Symm_CompatibilityError, VolumeError
from pyxtal.symmetry import Group
from pyxtal.symmetry import choose_wyckoff_mol as wyc_mol
from pyxtal.tolerance import Tol_matrix
from pyxtal.wyckoff_site import mol_site


# Define functions
# ------------------------------
class molecular_crystal:
    """
    Class for storing and generating molecular crystals based on symmetry
    constraints. Based on the crystal.random_crystal class for atomic crystals.
    Given a spacegroup, list of molecule objects, molecular stoichiometry, and
    a volume factor, generates a molecular crystal consistent with the given
    constraints.

    Args:
        dim: dimenion (1, 2, 3)
        group: the group number (1-75, 1-80, 1-230)
        molecules: a list of pymatgen.core.structure.Molecule objects for
            each type of molecule. Alternatively, you may supply a file path,
            or the name of molecules from the built_in
            `database <pyxtal.database.collection.html>`_
        numMols: A list of the number of each type of molecule within the
            primitive cell (NOT the conventioal cell)
        factor: A volume factor used to generate a larger or smaller
            unit cell. Increasing this gives extra space between molecules
        lattice (optional): the `Lattice <pyxtal.lattice.Lattice.html>`_
            object to define the unit cell
        conventional (optional): count the atomic numbers in a conventional cell
        tm (optional): the `Tol_matrix <pyxtal.tolerance.tolerance.html>`_
            object to define the distances
        sites (optional): pre-assigned wyckoff sites (e.g., `[["4a"], ["2b"]]`)
        seed (optional): seeds
        use_hall: False
    """

    def __init__(
        self,
        dim,
        group,
        molecules,
        numMols,
        factor=1.1,
        thickness=None,
        area=None,
        lattice=None,
        torsions=None,
        tm=Tol_matrix(prototype="molecular"),
        sites=None,
        conventional=True,
        seed=None,
        random_state=None,
        use_hall=False,
    ):
        # Initialize
        self.source = "Random"
        self.valid = False
        self.factor = factor
        self.seed = seed
        self.random_state = np.random.default_rng(random_state)

        # Dimesion
        self.dim = dim
        self.area = area  # Cross-section area for 1D
        self.thickness = thickness  # Thickness of 2D slab

        # The periodic boundary condition
        if dim == 3:
            self.PBC = [1, 1, 1]
        elif dim == 2:
            self.PBC = [1, 1, 0]
        elif dim == 1:
            self.PBC = [0, 0, 1]

        # Symmetry Group
        if type(group) == Group:
            self.group = group
        else:
            self.group = Group(group, dim=self.dim, use_hall=use_hall)
        self.number = self.group.number
        self.hall_number = self.group.hall_number

        # Composition: QZ: be cautious about the multiplicity
        if numMols is None:
            numMols = [len(self.group[0])] * len(molecules)
            no_check_compability = True
        else:
            numMols = np.array(numMols)  # must convert it to np.array
            no_check_compability = False
        mul = self.group.cellsize() if not conventional else 1
        self.numMols = numMols * mul

        # Tolerance matrix
        if type(tm) == Tol_matrix:
            self.tol_matrix = tm
        else:
            self.tol_matrix = Tol_matrix(prototype=tm)

        # Wyckofff sites
        self.set_molecules(molecules, torsions)
        self.set_sites(sites)
        valid_orientation = self.set_orientations()

        if not valid_orientation:
            msg = "Molecular symmetry is compatible with WP site\n"
            msg += "".join(f"{mol}: {mol.pga.sch_symbol}, " for mol in self.molecules)
            raise Symm_CompatibilityError(msg)

        # Check the minimum dof within the Wyckoff positions
        if no_check_compability:
            compat, self.has_freedom = True, True
        else:
            compat, self.has_freedom = self.group.check_compatible(self.numMols) #self.valid_orientations)

        if not compat:
            msg = f"Inincompatible compoisition {self.numMols} with symmetry {self.group.number}"
            raise Comp_CompatibilityError(msg)

        self.set_volume()
        self.set_lattice(lattice)
        self.set_crystal()

    def __str__(self):
        s = "------Random Molecular Crystal------"
        s += f"\nDimension: {self.dim}"
        # s += f"\nGroup: {self.symbol}"
        s += f"\nVolume factor: {self.factor}"
        s += f"\n{self.lattice}"
        if self.valid:
            s += "\nWyckoff sites:"
            s += "".join(f"\n\t{wyc}" for wyc in self.mol_sites)
        else:
            s += "\nStructure not generated."
        return s

    def __repr__(self):
        return str(self)

    def set_sites(self, sites):
        """
        Initialize and store symmetry sites for each molecule.

        This function processes a list of symmetry sites, validates them against the expected
        number of molecules, and stores them in the ``self.sites`` dictionary. Each entry in the
        ``sites`` list can be either a dictionary or another type (list, etc.), and if no valid
        site is provided for a molecule, `None` is assigned for that molecule's site.

        Args:
            sites (list): A list of sites corresponding to ``self.molecules``. They can be:
            - A dictionary of site information (keys represent Wyckoff letters or
                other identifiers, and values are the corresponding information).
            - A list or other type representing site information.
            - None, if no symmetry site information is available for that molecule.
        """
        # Initialize the self.sites dictionary to store site information
        self.sites = {}

        # Iterate over the molecules and their corresponding sites
        for i, _mol in enumerate(self.molecules):
            if sites is not None and sites[i] is not None and len(sites[i]) > 0:
                self._check_consistency(sites[i], self.numMols[i])
                if isinstance(sites[i], dict):
                    self.sites[i] = []
                    for item in sites[i].items():
                        self.sites[i].append({item[0]: item[1]})
                else:
                    self.sites[i] = sites[i]
            else:
                self.sites[i] = None

    def set_molecules(self, molecules, torsions):
        """
        Initialize and store molecular information.

        This function processes a list of molecules and initializes each one
        as a `pyxtal_molecule` object. If torsions are provided, they are
        applied to the respective molecules during initialization.
        It stored information in the `self.molecules` attribute.

        Args:
            molecules (list): A list of molecules, where each entry can either be:
                          - A SMILES string or file path representing a molecule.
                          - An already-initialized `pyxtal_molecule` object.
            torsions (list): A list of torsion angles. The length of `torsions`
                         must be equal to the length of `molecules`, or None
                         can be provided if no torsions are needed.
        """

        # If no torsions are provided, initialize with None for each molecule
        if torsions is None:
            torsions = [None] * len(molecules)

        # Initialize the molecules list to store processed pyxtal_molecule objects
        self.molecules = []

        # Iterate over the molecules
        for i, mol in enumerate(molecules):
            # already a pyxtal_molecule object
            if isinstance(mol, pyxtal_molecule):
                p_mol = mol
            else:
                p_mol = pyxtal_molecule(mol,
                                        seed=self.seed,
                                        torsions=torsions[i],
                                        tm=self.tol_matrix,
                                        random_state=self.random_state)
            self.molecules.append(p_mol)

    def set_orientations(self):
        """
        Calculates the valid orientations for each Molecule and Wyckoff
        position. Returns a list with 4 indices:
            - 1: the molecular prototype's index within self.molecules
            - 2: the WP's 1st index (based on multiplicity)
            - 3: the WP's 2nd index (within the group of same multiplicity)
            - 4: the index of a valid orientation for the molecule/WP pair

        For example, ``self.valid_orientations[i][j][k]`` would be a list of valid
        orientations for ``self.molecules[i]``, in the Wyckoff position
        ``self.group.wyckoffs_organized[j][k]``.
        """
        valid_ori = False
        self.valid_orientations = []
        for i, pyxtal_mol in enumerate(self.molecules):
            self.valid_orientations.append([])
            for x in self.group.wyckoffs_organized:
                self.valid_orientations[-1].append([])
                for _j, wp in enumerate(x):
                    # Don't check the wp with high multiplicity
                    if len(wp) > self.numMols[i]:
                        allowed = []
                    else:
                        allowed = pyxtal_mol.get_orientations_in_wp(wp)
                        if len(allowed) > 0:
                            valid_ori = True
                    self.valid_orientations[-1][-1].append(allowed)
        return valid_ori

    def set_volume(self):
        """
        Given the molecular stoichiometry, estimate the volume for a unit cell.
        """
        volume = 0
        for numMol, mol in zip(self.numMols, self.molecules):
            volume += numMol * mol.volume
        self.volume = abs(self.factor * volume)

    def set_lattice(self, lattice):
        """
        Generate the initial lattice.
        """
        if lattice is not None:
            # Use the provided lattice
            self.lattice = lattice
            self.volume = lattice.volume
            # Make sure the custom lattice PBC axes are correct.
            if lattice.PBC != self.PBC:
                self.lattice.PBC = self.PBC
                raise ValueError("PBC is incompatible " + str(self.PBC))
        else:
            # Determine the unique axis
            if self.dim == 2:
                unique_axis = "c" if self.number in range(3, 8) else "a"
            elif self.dim == 1:
                unique_axis = "a" if self.number in range(3, 8) else "c"
            else:
                unique_axis = "c"

            # Generate a Lattice instance
            good_lattice = False
            for _cycle in range(10):
                try:
                    if self.group.number < 10:
                        coef = 1.0 * self.numMols[0] / self.group[0].multiplicity
                    elif 10 <= self.group.number <= 15:
                        coef = 2.0  # 2/m
                    elif 16 <= self.group.number <= 74:
                        coef = 1.5
                    else:
                        coef = 1.0

                    self.lattice = Lattice(
                        self.group.lattice_type,
                        self.volume,
                        PBC=self.PBC,
                        unique_axis=unique_axis,
                        thickness=self.thickness,
                        area=self.area,
                        min_special=coef * max([mol.get_max_length() for mol in self.molecules]),
                        random_state=self.random_state,
                    )
                    good_lattice = True
                    break
                except VolumeError:
                    self.volume *= 1.1
                    msg = "Warning: increase the volume by 1.1 times: "
                    msg += f"{self.volume:.2f}"
                    print(msg)

            if not good_lattice:
                msg = f"Volume estimation {self.volume:.2f} is very bad"
                msg += " with the given composition "
                msg += str(self.numMols)
                raise RuntimeError(msg)

    def set_crystal(self):
        """
        The main code to generate a random molecular crystal.
        If successful, ``self.valid`` is True
        """
        self.numattempts = 0
        if not self.has_freedom:
            self.lattice_attempts = 20
            self.coord_attempts = 3
            self.ori_attempts = 1
        else:
            self.lattice_attempts = 40
            self.coord_attempts = 30
            self.ori_attempts = 5

        if not self.lattice.allow_volume_reset:
            self.lattice_attempts = 1

        for cycle1 in range(self.lattice_attempts):
            self.cycle1 = cycle1
            for cycle2 in range(self.coord_attempts):
                self.cycle2 = cycle2
                output = self._set_coords()

                if output:
                    self.mol_sites = output
                    break
            if self.valid:
                return
            else:
                self.lattice.reset_matrix()

        print("Cannot generate crystal after max attempts.")

    def _set_coords(self):
        """
        Generate coordinates for random crystal.
        """

        mol_sites_total = []
        # Add molecules
        for i, numMol in enumerate(self.numMols):
            pyxtal_mol = self.molecules[i]
            valid_ori = self.valid_orientations[i]
            output = self._set_mol_wyckoffs(i, numMol, pyxtal_mol, valid_ori, mol_sites_total)
            if output is not None:
                mol_sites_total.extend(output)
            else:
                # correct multiplicity not achieved exit and start over
                return None

        self.valid = True
        return mol_sites_total

    def _set_mol_wyckoffs(self, id, numMol, pyxtal_mol, valid_ori, mol_wyks):
        """
        Generates a set of wyckoff positions to accomodate a given number
        of molecules.

        Args:
            id: molecular id
            numMol: Number of ions to accomodate
            pyxtal_mol: Type of species being placed on wyckoff site
            valid_ori: list of valid orientations
            mol_wyks: current wyckoff sites

        Returns:
            if sucess, wyckoff_sites_tmp: list of wyckoff sites for valid sites
            otherwise, None

        """
        numMol_added = 0
        mol_sites_tmp = []

        # Now we start to add the specie to the wyckoff position
        sites_list = deepcopy(self.sites[id])  # the list of Wyckoff site
        if sites_list is not None:
            self.wyckoff_attempts = max(len(sites_list) * 2, 10)
        else:
            # the minimum numattempts is to put all atoms to the general WPs
            min_wyckoffs = int(numMol / len(self.group.wyckoffs_organized[0][0]))
            self.wyckoff_attempts = max(2 * min_wyckoffs, 10)

        for _cycle in range(self.wyckoff_attempts):
            # Choose a random WP for given multiplicity: 2a, 2b, 2c
            if sites_list is not None and len(sites_list) > 0:
                site = sites_list[0]
            else:  # Selecting the merging
                site = None

            # NOTE: The molecular version return wyckoff indices, not ops
            diff = numMol - numMol_added

            if type(site) is dict:  # site with coordinates
                key = next(iter(site.keys()))
                wp = wyc_mol(self.group, diff, key, valid_ori, True, self.dim, self.random_state)
            else:
                wp = wyc_mol(self.group, diff, site, valid_ori, True, self.dim, self.random_state)

            if wp is not False:
                # Generate a list of coords from the wyckoff position
                mult = wp.multiplicity  # remember the original multiplicity

                pt = site[key] if type(site) is dict else self.lattice.generate_point()
                # merge coordinates if the atoms are close
                mtol = pyxtal_mol.radius * 0.5
                pt, wp, oris = wp.merge(pt, self.lattice.matrix, mtol, valid_ori, self.group)

                if wp is not False:
                    if site is not None and mult != wp.multiplicity:
                        continue
                    if self.dim == 2 and self.thickness is not None and self.thickness < 0.1:
                        pt[-1] = 0.5

                    ms0 = self._set_orientation(pyxtal_mol, pt, oris, wp)
                    if not ms0.short_dist():
                        # Check current WP against existing WP's
                        passed_wp_check = True
                        # print("Checking", ms0)
                        for ms1 in mol_wyks + mol_sites_tmp:
                            if not ms0.short_dist_with_wp2(ms1, tm=self.tol_matrix):
                                passed_wp_check = False
                            # else:
                            #    print("passing", ms1)

                        if passed_wp_check:
                            if sites_list is not None:
                                sites_list.pop(0)

                            ms0.type = id
                            mol_sites_tmp.append(ms0)
                            numMol_added += len(ms0.wp)

                            # We have enough molecules of the current type
                            if numMol_added == numMol:
                                return mol_sites_tmp
        return None


    def _set_orientation(self, pyxtal_mol, pt, oris, wp):
        """
        Generate valid orientations for a given molecule in a Wyckoff position.

        It tries to generate valid orientations for the molecule by:
            - Selecting a random orientation from a list of possible orientations.
            - Flipping the orientation to test different alignments.
            - Checking the smallest distance between atoms and ensuring it's valid.
            - Using the bisection method is refine the orientation.

        Args:
            pyxtal_mol: The pyxtal_molecule object representing the molecule.
            pt: Position of the molecule.
            oris: List of potential orientations.
            wp: Wyckoff position object representing the symmetry of the site.

        Returns:
            ms0: A valid ``mol_site`` object if an acceptable orientation is found.
             returns `None` if no valid orientation is found within the attempts.
        """

        # NOTE removing this copy causes tests to fail -> state not managed well
        ori = self.random_state.choice(oris).copy()
        ori.change_orientation(flip=True)

        # Create a mol_site object with the current orientation
        ms0 = mol_site(pyxtal_mol, pt, ori, wp, self.lattice)

        # Check if the current orientation results in valid distances
        if len(pyxtal_mol.mol) > 1 and ori.degrees > 0:
            if ms0.short_dist():
                ms0.optimize_orientation_by_dist(self.ori_attempts)
        return ms0

    def _check_consistency(self, site, numMol):
        """
        Check if N_mol is consistent with the symmetry constraints of the system.

        It verifies if the sum of molecules from the WP matches (`numMol`).
        Each Wyckoff site string in the `site` list includes a number that
        represents how many molecules are associated with that site.

        If a inconsistency is found, it raises a ValueError with a detailed message.

        Args:
            site (list of str): A list of strings for Wyckoff sites. Each string
                            contains a number (e.g., "3a", "4b") where the number
                            indicates how many molecules are at that site.
            numMol (int): The total number of molecules expected in the structure.

        Returns:
            bool: Returns `True` if the number of molecules matches `numMol`.
                Raises a ValueError if they do not match.
        """
        num = 0
        for s in site:
            num += int(s[:-1])
        if numMol == num:
            return True
        else:
            msg = "\nThe requested number of molecules is inconsistent: "
            msg += str(site)
            msg += f"\nfrom numMols: {numMol:d}"
            msg += f"\nfrom Wyckoff list: {num:d}"
            raise ValueError(msg)



from pyxtal.symmetry import Group
import csv

def csv_to_rst(csv_path, rst_path):
    with open(csv_path, 'r') as csvfile, open(rst_path, 'w') as rstfile:
        # Write RST header
        rstfile.write("Wyckoff Positions\n")
        rstfile.write("=================\n\n")

        # Read CSV
        reader = csv.reader(csvfile)
        headers = next(reader)

        # Write table header
        rstfile.write(".. list-table::\n")
        rstfile.write("   :header-rows: 1\n")
        rstfile.write("   :widths: auto\n\n")

        # Write headers
        rstfile.write("   * - " + "\n     - ".join(headers) + "\n")

        # Write data rows
        for row in reader:
            rstfile.write("   * - " + "\n     - ".join(row) + "\n")



# Open a CSV file for writing
with open('doc/wyckoff_positions.csv', 'w', newline='') as csvfile:
    # Create CSV writer
    csvwriter = csv.writer(csvfile)

    # Write header
    csvwriter.writerow(['Space Group Number', 'Space Group Symbol', 'Wyckoff Label', 'Site Symmetry'])

    # Iterate through space groups
    for g in range(1, 231):
        spg = Group(g)
        for wp in spg:
            wp.get_site_symmetry()
            # Write data row
            csvwriter.writerow([spg.number, spg.symbol, wp.get_label(), wp.site_symm])

# Convert CSV to RST
csv_to_rst('doc/wyckoff_positions.csv', 'doc/wyckoff_positions.rst')



"""
This module handles visualization. Mostly powered by py3Dmol
"""


def addBox(view, vecs, viewer=None):
    pts = [
        ([0, 0, 0], vecs[0, :]),
        ([0, 0, 0], vecs[1, :]),
        ([0, 0, 0], vecs[2, :]),
        (vecs[0, :], vecs[0, :] + vecs[1, :]),
        (vecs[0, :], vecs[0, :] + vecs[2, :]),
        (vecs[1, :], vecs[0, :] + vecs[1, :]),
        (vecs[1, :], vecs[2, :] + vecs[1, :]),
        (vecs[2, :], vecs[0, :] + vecs[2, :]),
        (vecs[2, :], vecs[1, :] + vecs[2, :]),
        (vecs[0, :] + vecs[1, :], vecs[0, :] + vecs[1, :] + vecs[2, :]),
        (vecs[0, :] + vecs[2, :], vecs[0, :] + vecs[1, :] + vecs[2, :]),
        (vecs[1, :] + vecs[2, :], vecs[0, :] + vecs[1, :] + vecs[2, :]),
    ]

    for _i, pt in enumerate(pts):
        pt1, pt2 = pt

        if viewer is None:
            view.addLine(
                {
                    "start": {"x": pt1[0], "y": pt1[1], "z": pt1[2]},
                    "end": {"x": pt2[0], "y": pt2[1], "z": pt2[2]},
                }
            )
        else:
            view.addLine(
                {
                    "start": {"x": pt1[0], "y": pt1[1], "z": pt1[2]},
                    "end": {"x": pt2[0], "y": pt2[1], "z": pt2[2]},
                },
                viewer=viewer,
            )

    # add unitcell labels
    labels = {"o": [0, 0, 0], "a": vecs[0, :], "b": vecs[1, :], "c": vecs[2, :]}
    for key in labels:
        text, pos = key, labels[key]
        if viewer is None:
            view.addLabel(
                text,
                {
                    "position": {"x": pos[0], "y": pos[1], "z": pos[2]},
                    "fontColor": "black",
                    "backgroundColor": "white",
                    "fontsize": 12,
                    "backgroundOpacity": "0.1",
                },
            )
        else:
            view.addLabel(
                text,
                {
                    "position": {"x": pos[0], "y": pos[1], "z": pos[2]},
                    "fontColor": "black",
                    "backgroundColor": "white",
                    "fontsize": 12,
                    "backgroundOpacity": "0.1",
                },
                viewer=viewer,
            )


def addlines(view, orig, axes, viewer=None):
    for ax in axes:
        ax += orig
        if viewer is None:
            view.addArrow(
                {
                    "start": {"x": orig[0], "y": orig[1], "z": orig[2]},
                    "end": {"x": ax[0], "y": ax[1], "z": ax[2]},
                }
            )
        else:
            view.addArrow(
                {
                    "start": {"x": orig[0], "y": orig[1], "z": orig[2]},
                    "end": {"x": ax[0], "y": ax[1], "z": ax[2]},
                },
                viewer=viewer,
            )


def display_atomic(struc, size=(600, 300), scale=0.25, radius=0.10, supercell=(1, 1, 1), show_wp=False):
    """
    display the molecular crystals generated from pyxtal. If the animation is False,
    only dump the structure to cif and let py3Dmol display it. If animation is on,
    show the generation of molecules in the crystal as an animation.

    Args:
        size: (width, height) in tuple
        scale: the size of sphere
        radius: the size of stick
        supercell: replicate the crystal (valid only when animation is False)
        show_wp: whether or not highlight the unique wyckoff site

    Returns:
        py3Dmol object
    """
    import py3Dmol

    (width, height) = size
    view = py3Dmol.view(height=height, width=width)
    fmt = "xyz" if struc.dim == 0 else "cif"

    txt = struc.to_file(fmt=fmt)
    view.addModel(txt, fmt, {"doAssembly": True, "duplicateAssemblyAtoms": True})
    view.setStyle(
        {
            "sphere": {"colorscheme": "Jmol", "scale": scale},
            "stick": {"colorscheme": "Jmol", "radius": radius},
        }
    )
    if struc.dim != 0:
        view.addUnitCell()
        A, B, C = supercell
        view.replicateUnitCell(A, B, C)
        if show_wp:
            view.setStyle({"sym": 2}, {"sphere": {"scale": scale * 1.1, "color": "blue"}})
    return view.zoomTo()


def display_molecular(
    struc,
    size=(600, 300),
    supercell=(1, 1, 1),
    axis=None,
    animation=False,
    interval=2000,
):
    """
    display the molecular crystals generated from pyxtal. If the animation is False,
    only dump the structure to cif and let py3Dmol display it. If animation is on,
    show the generation of molecules in the crystal as an animation.

    Args:
        size: (width, height) in tuple
        supercell: replicate the crystal (valid only when animation is False)
        axis: 3-vector to reprent the rotational axis
        animation: whether or not display the animation

    Returns:
        py3Dmol object
    """
    import py3Dmol

    (width, height) = size
    view = py3Dmol.view(height=height, width=width)
    if not animation:
        cif = struc.to_file()
        view.addModel(
            cif,
            "cif",
            {
                "doAssembly": True,
                "duplicateAssemblyAtoms": True,
                "normalizeAssembly": True,
            },
        )
        view.setStyle({"stick": {"colorscheme": "greenCarbon"}})
        view.addUnitCell()
        if axis is not None:
            site = struc.mol_sites[0]
            for id in range(len(site.wp)):
                mol = site.get_mol_object(id)
                addlines(view, mol.cart_coords.mean(axis=0), [axis.copy()])

        A, B, C = supercell
        view.replicateUnitCell(A, B, C)
    else:
        cifs = ""
        for i in range(len(struc.mol_sites[0].wp)):
            cifs += struc.to_file(sym_num=i + 1)
        view.addModelsAsFrames(cifs, "cif")
        view.animate({"loop": "forward", "interval": interval})
        view.addUnitCell()
        view.setStyle({"model": 0}, {"stick": {"colorscheme": "greenCarbon"}})
    return view.zoomTo()


def display_molecular_site(site, id=None, size=(400, 300), axis=True, ax_id=range(3), box=False):
    """
    display the Wyckoff site in the molecular crystals generated from pyxtal.

    Args:
        site: pyxtal.wyckoff_site.mol_site object
        id: list of molecules to display. If None, display all molecules in this site
        size: (width, height) in tuple
        axis: boolean, whether or not display the rotational axis
        axis_label: whether or not display the unitcell axis labels

    Returns:
        py3Dmol object
    """
    import py3Dmol

    (width, height) = size
    view = py3Dmol.view(height=height, width=width)

    ids = range(site.wp.multiplicity) if id is None else [id]

    for i in ids:
        mol = site.get_mol_object(i)
        cell, _, center = site.molecule.get_box_coordinates(mol.cart_coords)
        view.addModel(mol.to(fmt="xyz"), "xyz")
        if axis:
            axes = site.molecule.get_principle_axes(mol.cart_coords)
            addlines(view, mol.cart_coords.mean(axis=0), axes.T[ax_id] * 5, viewer=(0, 1))
        if box:
            center_spec = {"x": center[0], "y": center[1], "z": center[2]}
            w_spec = {"x": cell[0, 0], "y": cell[0, 1], "z": cell[0, 2]}
            h_spec = {"x": cell[1, 0], "y": cell[1, 1], "z": cell[1, 2]}
            d_spec = {"x": cell[2, 0], "y": cell[2, 1], "z": cell[2, 2]}
            view.addBox(
                {
                    "center": center_spec,
                    "dimensions": {"w": w_spec, "h": h_spec, "d": d_spec},
                    "color": "magenta",
                    "alpha": 0.5,
                }
            )

    addBox(view, site.lattice.matrix, viewer=(0, 1))
    view.setStyle({"stick": {"colorscheme": "greenCarbon"}})

    return view.zoomTo()


def display_molecules(molecules, size=(400, 300), animation=False, center=False):
    """
    display the molecules in Pymatgen object.

    Args:
        molecules: a list of pymatgen molecules
        size: (width, height) in tuple
        animation: whether or not display animation
        center: highlight center or not

    Returns:
        py3Dmol object

    """
    import py3Dmol

    (width, height) = size
    view = py3Dmol.view(height=height, width=width)
    mol_strs = ""
    for mol in molecules:
        mol_strs += mol.to(fmt="xyz") + "\n"
    if animation:
        view.addModelsasFrames(mol_strs, "xyz")
        view.animate({"loop": "forward"})
    else:
        view.addModels(mol_strs, "xyz")
    view.setStyle({"stick": {"colorscheme": "greenCarbon"}})

    return view.zoomTo()


def display_molecule(molecule, center, cell, size=(400, 300)):
    """
    display the molecules in Pymatgen object.

    Args:
        mol: pymatgen molecule
        center: molecular
        size: (width, height) in tuple

    Returns:
        py3Dmol object

    """
    import py3Dmol

    (width, height) = size
    view = py3Dmol.view(height=height, width=width)
    mol_strs = ""
    mol_strs += molecule.to(fmt="xyz") + "\n"
    view.addModels(mol_strs, "xyz")
    center_spec = {"x": center[0], "y": center[1], "z": center[2]}
    w_spec = {"x": cell[0, 0], "y": cell[0, 1], "z": cell[0, 2]}
    h_spec = {"x": cell[1, 0], "y": cell[1, 1], "z": cell[1, 2]}
    d_spec = {"x": cell[2, 0], "y": cell[2, 1], "z": cell[2, 2]}
    view.addBox(
        {
            "center": center_spec,
            "dimensions": {"w": w_spec, "h": h_spec, "d": d_spec},
            "color": "magenta",
            "opacity": 0.4,
            #'alpha': 0.6,
        }
    )
    view.setStyle({"stick": {"colorscheme": "greenCarbon"}})

    return view.zoomTo()


def display_mol_crystals(
    strucs,
    size=(600, 300),
    supercell=(1, 1, 1),
    axis=None,
    animation="slider",
    interval=2000,
):
    """
    display the molecular crystals generated from pyxtal.
    two modes of animations are supported: slider or movie

    Args:
        size: (width, height) in tuple
        supercell: replicate the crystal (valid only when animation is False)
        animation: slider or movie

    Returns:
        py3Dmol object
    """
    import py3Dmol

    (width, height) = size
    view = py3Dmol.view(height=height, width=width)
    if animation == "slider":
        from ipywidgets import IntSlider, interact

        def conf_viewer(idx):
            return strucs[idx].show(size=size, supercell=supercell, axis=axis)

        interact(conf_viewer, idx=IntSlider(min=0, max=len(strucs) - 1, description="id:"))
        return None

    elif animation == "movie":
        cifs = ""
        for struc in strucs:
            cifs += struc.to_file()
        view.addModelsAsFrames(cifs, "cif")
        view.animate({"loop": "forward", "interval": interval})
        view.addUnitCell()
        view.setStyle({"model": 0}, {"stick": {"colorscheme": "greenCarbon"}})
        return view.zoomTo()
    else:
        raise ValueError("only movie or slider is supported")


def display_crystals(strucs, size=(600, 300), supercell=(1, 1, 1), labels=None):
    """
    display the molecular crystals generated from pyxtal.
    two modes of animations are supported: slider or movie

    Args:
        size: (width, height) in tuple
        supercell: replicate the crystal (valid only when animation is False)
        labels: labels for each structures

    Returns:
        py3Dmol object
    """
    import py3Dmol
    from ipywidgets import IntSlider, interact

    (width, height) = size
    py3Dmol.view(height=height, width=width)
    if labels is None:
        _l_min, _l_max = 0, len(strucs) - 1
    else:
        _l_min, _l_max = labels[0], labels[-1]

    def conf_viewer(idx):
        return strucs[idx].show(size=size, supercell=supercell)

    interact(conf_viewer, idx=IntSlider(min=0, max=len(strucs) - 1, description="id:"))


def display_cluster(molecules, cell, Ps, cmap="YlGn", s_opacity=0.5, size=(400, 300), style="sphere"):
    import numpy as np
    import py3Dmol
    from matplotlib import cm, colors

    cmaps = getattr(cm, cmap)(Ps / np.mean(Ps))
    models = {}

    view = py3Dmol.view()
    mol_str = ""
    mol_str += molecules[0].to(fmt="xyz") + "\n"
    view.addModel(mol_str, "xyz")
    model = view.getModel()
    model.setStyle({}, {"sphere": {"colorscheme": "grayCarbon", "scale": 0.7}})

    center = molecules[0].center_of_mass
    {"x": center[0], "y": center[1], "z": center[2]}
    {"x": cell[0, 0], "y": cell[0, 1], "z": cell[0, 2]}
    {"x": cell[1, 0], "y": cell[1, 1], "z": cell[1, 2]}
    {"x": cell[2, 0], "y": cell[2, 1], "z": cell[2, 2]}
    # view.addBox({'center': center_spec,
    #             'dimensions': {'w': w_spec, 'h': h_spec, 'd': d_spec},
    #             'color':'magenta',
    #             'alpha': 0.5,
    #             #'opacity': 0.1,
    #            })
    addBox(view, cell)

    for i in range(1, len(molecules)):
        mol = molecules[i]
        mol_strs = ""
        mol_strs += mol.to(fmt="xyz") + "\n"
        view.addModel(mol_strs, "xyz")
        model = view.getModel()
        color = "greenCarbon"
        opacity = 0.65
        if style == "sphere":
            model.setStyle({}, {"sphere": {"colorscheme": color, "scale": 0.5, "opacity": opacity}})
        else:
            model.setStyle({}, {"stick": {"colorscheme": color, "radius": 0.1, "opacity": opacity}})

        view.addSurface(py3Dmol.VDW, {"opacity": s_opacity, "color": colors.rgb2hex(cmaps[i - 1])})

    return view.zoomTo({"model": list(models.values())})
